
function ChannelSelection(N,seed)

rng(seed);


%N = 4; % Number of channels

for i=1:N
   a(i)=rand();
   b(i)=a(i)+(1-a(i)).*rand();
   %b(i)=1-(1-a(i)).*rand();
   %b(i)=1;
end

% Setup 1 channels
a(1)=0.6;
b(1)=0.8;
a(2)=0.2;
b(2)=0.4;
a(3)=0;
b(3)=1;
a(4)=0.3;
b(4)=0.7;

% Setup 2 channels
% a(1)=0.5;
% b(1)=0.8;
% a(2)=0.4;
% b(2)=0.9;
% a(3)=0.5;
% b(3)=0.8;
% a(4)=0.4;
% b(4)=0.6;

disp('Channel Parameters');
disp([a' b']);

% -----------------

TimeHorizon=100;

%----------------- Brute force ------------------

M=1;
Q_BF = zeros(1,N);
R=zeros(1,TimeHorizon);
ch_selected = zeros(1,TimeHorizon);

% I check all the channels

time_to_check_all = 1;

for n=1:N
    for m=1:M
        ch_selected(time_to_check_all) = n; 
        rho = ChannelOccupancy(a(n),b(n));
        R(time_to_check_all) = 1-rho;
        Q_BF(n)=Q_BF(n) + (1-rho);
        time_to_check_all = time_to_check_all + 1;
    end
    Q_BF(n)=Q_BF(n)/M;
end
   
% I pick the one with highest reward
    
disp(Q_BF);

[~,n_best]=max(Q_BF);
    
for t=time_to_check_all:TimeHorizon
    ch_selected(t) = n_best; 
    rho = ChannelOccupancy(a(n_best),b(n_best));
    R(t) = 1-rho;
end

AccR = cumsum(R);

figure
yyaxis left
plot(ch_selected);
ylabel('Channel selected (Brute Force)');
axis([0 TimeHorizon 0 N+1]);

yyaxis right
plot(AccR./TimeHorizon);
hold
plot(R);
hold
ylabel('Reward and Accumulated Reward');
axis([0 TimeHorizon 0 1]);
xlabel('Time');

%----------------- Epsilon greedy ------------------

epsilon = 0.1;
Delta = 0/(4*N);

Q_EG = zeros(1,N);
R_EG = zeros(1,TimeHorizon);
ch_selected_EG = zeros(1,TimeHorizon);
n_times_ch_selected_EG = zeros(1,N);

% First Iteration

n_sel = 1;

ch_selected_EG(1) = n_sel; 
rho = ChannelOccupancy(a(n_sel),b(n_sel));
R(1) = 1-rho;
Q_EG(n_sel)=Q_EG(n_sel) + (1-rho);

n_times_ch_selected_EG(n_sel) = 1;

for t=2:TimeHorizon
   
    if(rand < epsilon) % explore
        n_sel=randi(N,1);
    else % exploit
        [~,n_sel]=max(Q_EG);
    end
    
    ch_selected_EG(t) = n_sel;
    n_times_ch_selected_EG(n_sel)=n_times_ch_selected_EG(n_sel)+1;
    rho = ChannelOccupancy(a(n_sel),b(n_sel));
    R_EG(t) = 1-rho;
    Q_EG(n_sel)=(Q_EG(n_sel)*(n_times_ch_selected_EG(n_sel)-1) + (1-rho))/n_times_ch_selected_EG(n_sel);
    
    epsilon = epsilon - Delta;
end

AccR_EG = cumsum(R_EG);


figure
yyaxis left
plot(ch_selected_EG);
ylabel('Channel selected (\epsilon-greedy)');
axis([0 TimeHorizon 0 N+1]);

yyaxis right
plot(AccR_EG./TimeHorizon);
hold
plot(R_EG);
hold
ylabel('Reward and Accumulated Reward');
axis([0 TimeHorizon 0 1]);
xlabel('Time');


%----------------- Thompson Sampling ------------------


Q_TS = zeros(1,N);
R_TS = zeros(1,TimeHorizon);
ch_selected_TS = zeros(1,TimeHorizon);
n_times_ch_selected_TS = zeros(1,N);

% First Iteration

n_sel = 1;

ch_selected_TS(1) = n_sel; 
rho = ChannelOccupancy(a(n_sel),b(n_sel));
R(1) = 1-rho;
Q_TS(n_sel)=Q_TS(n_sel) + (1-rho);
n_times_ch_selected_TS(n_sel) = 1;

sigm = zeros(1,N);


for t=2:TimeHorizon
   
    Sampling_arms=zeros(1,N);
    for i=1:N
        m = Q_TS(i);
        sigm(i) = (1/(n_times_ch_selected_TS(i)+1));
        Sampling_arms(i)=m + sigm(i).*randn();
    end
      
    [~,n_sel]=max(Sampling_arms);
        
    ch_selected_TS(t) = n_sel;
    n_times_ch_selected_TS(n_sel)=n_times_ch_selected_TS(n_sel)+1;
    rho = ChannelOccupancy(a(n_sel),b(n_sel));
    R_TS(t) = 1-rho;
    Q_TS(n_sel)=(Q_TS(n_sel)*(n_times_ch_selected_TS(n_sel)-1) + (1-rho))/n_times_ch_selected_TS(n_sel);

end

AccR_TS = cumsum(R_TS);


figure
yyaxis left
plot(ch_selected_TS);
ylabel('Channel selected (Thompson Sampling)');
axis([0 TimeHorizon 0 N+1]);

yyaxis right
plot(AccR_TS./TimeHorizon);
hold
plot(R_TS);
hold
ylabel('Reward and Accumulated Reward');
axis([0 TimeHorizon 0 1]);
xlabel('Time');


if(N==4)
x = [0:0.01:1]; 
y1 = normpdf(x,Q_TS(1),sigm(1));
y2 = normpdf(x,Q_TS(2),sigm(2));
y3 = normpdf(x,Q_TS(3),sigm(3));
y4 = normpdf(x,Q_TS(4),sigm(4));


figure
plot(x,y1./max(y1));
hold
plot(x,y2./max(y2));
plot(x,y3./max(y3));
plot(x,y4./max(y4));
hold
xlabel('values');
ylabel('pdf');
legend('Channel 1','Channel 2','Channel 3','Channel 4');


end

%----------------- UCB ------------------
  
Q_UCB = zeros(1,N);
R_UCB = zeros(1,TimeHorizon);
ch_selected_UCB = zeros(1,TimeHorizon);
n_times_ch_selected_UCB = zeros(1,N);

UCBx = N;

% First Iteration

for i=1:N
    n_sel = i;
    ch_selected_UCB(n_sel) = n_sel; 
    rho = ChannelOccupancy(a(n_sel),b(n_sel));
    R(i) = 1-rho;
    Q_UCB(n_sel)=Q_UCB(n_sel) + (1-rho);
    n_times_ch_selected_UCB(n_sel) = 1;
end

for t=N+1:TimeHorizon
   
    Sampling_arms=zeros(1,N);
    for i=1:N
        m = Q_UCB(i);
        Sampling_arms(i)= UCBx*m + sqrt(2*log(t)/n_times_ch_selected_UCB(i));
    end
        
    [~,n_sel]=max(Sampling_arms);
            
    ch_selected_UCB(t) = n_sel;
    n_times_ch_selected_UCB(n_sel)=n_times_ch_selected_UCB(n_sel)+1;
    rho = ChannelOccupancy(a(n_sel),b(n_sel));
    R_UCB(t) = 1-rho;
    Q_UCB(n_sel)=(Q_UCB(n_sel)*(n_times_ch_selected_UCB(n_sel)-1) + (1-rho))/n_times_ch_selected_UCB(n_sel);

end

AccR_UCB = cumsum(R_UCB);


figure
yyaxis left
plot(ch_selected_UCB);
ylabel('Channel selected (UCB)');
axis([0 TimeHorizon 0 N+1]);

yyaxis right
plot(AccR_UCB./TimeHorizon);
hold
plot(R_UCB);
hold
ylabel('Reward and Accumulated Reward');
axis([0 TimeHorizon 0 1]);
xlabel('Time');


% -------------------------------
disp('Which is the best one? Brute Force (M=1) | EG | TS | UCB');
disp([AccR(TimeHorizon) AccR_EG(TimeHorizon) AccR_TS(TimeHorizon) AccR_UCB(TimeHorizon)]);


end

function rho = ChannelOccupancy(a,b)

rho = a + (b-a).*rand();

end